#include "bpop_main.h"

GameScene currentScene = GAME_SCENE_MENU;
GameScene nextScene = GAME_SCENE_NONE;

GameDifficulty gameDifficulty = GAME_DIFFICULTY_NORMAL;

int outlineSizeLoc = 0;
int outlineColorLoc = 0;
int textureSizeLoc = 0;

Camera3D camera = { 
    .position = { 0.0f, 0.0f, -10.0f },
    .target = { 0.0f, 0.0f, 0.0f },
    .up = { 0.0f, 1.0f, 0.0f },
    .projection = CAMERA_ORTHOGRAPHIC,
    .fovy = 350.0f,
};


Playfield playfield = {
    .fieldSize = {250, 300},
    .waterLevel = 200.0f,
    .minHeight = 90.0f,
    .maxHeight = 280.0f,
};

GameTime gameTime = {
    .fixedDeltaTime = 1.0f / 60.0f,
};

Color bubbleTypeColors[] = {
    COLOR_FROM_HEX(0xaaccff),
    COLOR_FROM_HEX(0xaa7733),
    COLOR_FROM_HEX(0x33ffaa),
    COLOR_FROM_HEX(0xff33aa),
    COLOR_FROM_HEX(0xaa33ff),
};

Model bubbleModel;
Shader bubbleOutlineShader;
RenderTexture2D bubbleFieldTexture;

float BubbleLevelRadius(int level)
{
    return powf((level + 1) * 20, .75f);
}

int Button(const char *text, Vector2 position, Vector2 size)
{
    int result = 0;
    Rectangle rect = {position.x, position.y, size.x, size.y};
    if (CheckCollisionPointRec(GetMousePosition(), rect))
    {
        if (IsMouseButtonReleased(MOUSE_LEFT_BUTTON))
        {
            result = 1;
        }
    }
    DrawRectangleRec(rect, (Color){200, 200, 200, 255});
    DrawRectangleLinesEx(rect, 2, BLACK);
    float width = MeasureText(text, 20);
    DrawText(text, position.x + size.x * 0.5f - width * 0.5f, position.y + 10, 20, BLACK);
    return result;
}

void PlayfieldFixedUpdate(Playfield *playfield)
{
    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        bubble->sameLevelContact = 0;
    }

    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        if (!bubble->flagIsActive) continue;
        float r = bubble->radius;

        for (int j = i + 1; j < MAX_BUBBLES; j++)
        {
            Bubble *other = &playfield->bubbles[j];
            if (!other->flagIsActive) continue;
            float otherR = other->radius;
            float sumR2 = (r + otherR) * (r + otherR);
            float d2 = Vector2DistanceSqr(bubble->position, other->position);
            int canMerge = bubble->bubbleLevel == other->bubbleLevel && bubble->bubbleType == other->bubbleType;
            if (d2 < sumR2 * 1.05f)
            {
                if (canMerge)
                {
                    bubble->sameLevelContact = 1;
                    other->sameLevelContact = 1;
                }
                if (canMerge && bubble->bubbleMergeCooldown <= 0.0f 
                    && other->bubbleMergeCooldown <= 0.0f)
                {
                    // merge bubbles
                    bubble->bubbleLevel++;
                    bubble->radius = BubbleLevelRadius(bubble->bubbleLevel);
                    bubble->bubbleMergeCooldown = 1.0f;
                    other->flagIsActive = 0;
                }
            }

            if (d2 < sumR2)
            {
                float overlap = r + otherR - sqrtf(d2);
                Vector2 normal = Vector2Normalize(Vector2Subtract(other->position, bubble->position));
                // resolve overlap by moving the bubbles apart
                const float errorCorrection = 0.25f;
                bubble->position = Vector2Subtract(bubble->position, Vector2Scale(normal, overlap * errorCorrection));
                other->position = Vector2Add(other->position, Vector2Scale(normal, overlap * errorCorrection));

                // bounce off each other
                Vector2 relativeVelocity = Vector2Subtract(bubble->velocity, other->velocity);
                float dot = Vector2DotProduct(relativeVelocity, normal);
                if (dot > 0.0f)
                {
                    // DrawLineV(bubble->position, other->position, COLOR_FROM_HEX(0xff0000));
                    float impulse = -dot * 0.85f;
                    bubble->velocity = Vector2Add(bubble->velocity, Vector2Scale(normal, impulse));
                    other->velocity = Vector2Subtract(other->velocity, Vector2Scale(normal, impulse));
                }
            }
        }

        if (!bubble->sameLevelContact)
        {
            bubble->bubbleMergeCooldown = 1.0f;
        }
        else
        {
            bubble->bubbleMergeCooldown -= gameTime.fixedDeltaTime;
        }

        float buoyancy = -20.0f;
        if (bubble->position.y < playfield->waterLevel)
        {
            buoyancy = (playfield->waterLevel - bubble->position.y) * 0.5f;
        }

        int isOutsideZone = bubble->position.y < playfield->minHeight + bubble->radius || 
            bubble->position.y > playfield->maxHeight - bubble->radius;
        bubble->lifeTime += gameTime.fixedDeltaTime;
        bubble->isOutsideZone = isOutsideZone && bubble->lifeTime > 1.0f;

        bubble->velocity = Vector2Add(bubble->velocity, (Vector2){0, buoyancy});
        bubble->velocity = Vector2Scale(bubble->velocity, 0.92f);
        bubble->position.x += bubble->velocity.x * gameTime.fixedDeltaTime;
        bubble->position.y += bubble->velocity.y * gameTime.fixedDeltaTime;
        if ((bubble->position.x < r && bubble->velocity.x < 0.0f) || 
            (bubble->position.x > playfield->fieldSize.x - r && bubble->velocity.x > 0.0f))
        {
            bubble->velocity.x *= -0.9f;
        }

        bubble->position.x = (bubble->position.x < r) ? r : (bubble->position.x > playfield->fieldSize.x - r) ? playfield->fieldSize.x - r : bubble->position.x;
        // bubble->position.y = (bubble->position.y < r) ? r : (bubble->position.y > playfield->fieldSize.y - r) ? playfield->fieldSize.y - r : bubble->position.y;

        // debug velocity
        // DrawLineV(bubble->position, Vector2Add(bubble->position, Vector2Scale(bubble->velocity, 1.0f)), COLOR_FROM_HEX(0xff0000));
    }

    int outsideCount = 0;
    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        if (bubble->isOutsideZone)
        {
            outsideCount++;
        }
    }

    if (outsideCount == 0)
    {
        playfield->strikes = 0;
        playfield->gameoverCooldown = 0.0f;
    }
    if (playfield->strikes >= MAX_STRIKES)
    {
        playfield->gameoverCooldown += gameTime.fixedDeltaTime;
        if (playfield->gameoverCooldown > GAMEOVER_COOLDOWN)
        {
            nextScene = GAME_SCENE_GAMEOVER;
        }
    }
}

void PlayfieldTryAddBubble(Playfield *playfield, Vector2 position)
{
    if (playfield->strikes >= MAX_STRIKES) return;

    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        if (!bubble->flagIsActive)
        {
            bubble->flagIsActive = 1;
            bubble->position = position;
            bubble->velocity = (Vector2){GetRandomValue(-100, 100), GetRandomValue(-100, 100)};
            bubble->bubbleType = playfield->nextBubbleType;
            bubble->bubbleLevel = playfield->nextBubbleLevel;
            bubble->radius = BubbleLevelRadius(bubble->bubbleLevel);
            bubble->lifeTime = 0.0f;
            playfield->spawnedBubbleCount++;

            for (int j = 0; j < MAX_BUBBLES; j += 1)
            {
                Bubble *other = &playfield->bubbles[j];
                if (!other->flagIsActive) continue;
                if (other->isOutsideZone)
                {
                    playfield->strikes++;
                    break;
                }
            }

            playfield->nextBubbleType = GetRandomValue(0, gameDifficulty);
            playfield->nextBubbleLevel = GetRandomValue(0, 3);
            break;
        }
    }

}

PlayfieldScores CalculatePlayfieldScores(Playfield *playfield)
{
    PlayfieldScores scores = {0};
    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        if (bubble->flagIsActive)
        {
            scores.bubbleCount++;
            uint32_t bubbleScore = 1 << bubble->bubbleLevel;
            scores.score += bubbleScore;

            if (bubble->isOutsideZone)
            {
                scores.outsideBubbleCount++;
            }
        }
    }
    scores.score += playfield->spawnedBubbleCount;
    return scores;
}

Vector2 PlayfieldPositionToSpawnPosition(Playfield *playfield, Vector2 position)
{
    Vector2 spawnPosition = position;
    spawnPosition.y = BubbleLevelRadius(5);
    return spawnPosition;
}

Vector2 PlayfieldScreenToSpawnPosition(Playfield *playfield, Camera3D camera, Vector2 screenPosition)
{
    Vector3 cursorPosition = GetScreenToWorldRay(screenPosition, camera).position;
    cursorPosition.x += playfield->fieldSize.x / 2;
    cursorPosition.y += playfield->fieldSize.y / 2;

    Vector2 pos = {cursorPosition.x, cursorPosition.y};
    return PlayfieldPositionToSpawnPosition(playfield, pos);
}

void PlayfieldDrawBubbles(Playfield *playfield, Camera3D camera)
{
    float bubbleExtraRadius = 5.0f;
    Vector2 mousePos = GetMousePosition();
    Vector2 spawnPosition = PlayfieldScreenToSpawnPosition(playfield, camera, mousePos);
    DrawCube((Vector3){0, 0, 0}, playfield->fieldSize.x, playfield->fieldSize.y, 0, COLOR_FROM_HEX(0xbbddff));
    DrawCube((Vector3){0, (playfield->waterLevel - playfield->fieldSize.y) * 0.5f, 0}, playfield->fieldSize.x, 
        playfield->waterLevel, 0, COLOR_FROM_HEX(0x225588));

    DrawModel(bubbleModel, 
        (Vector3){spawnPosition.x - playfield->fieldSize.x * 0.5f, spawnPosition.y - playfield->fieldSize.y * 0.5f, 0},
        BubbleLevelRadius(playfield->nextBubbleLevel) + bubbleExtraRadius, bubbleTypeColors[playfield->nextBubbleType]);
    // DrawRectangle(0, 0, playfield->fieldSize.x, playfield->fieldSize.y, COLOR_FROM_HEX(0x225588));
    rlPushMatrix();
    rlTranslatef(-playfield->fieldSize.x / 2, -playfield->fieldSize.y / 2, 0);
    // draw bubbles into playfield space
    float blink = sinf(gameTime.time * 10.0f) * 0.2f + 0.5f;
    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        if (!bubble->flagIsActive) continue;
        Vector3 position = {bubble->position.x, bubble->position.y, 0};
        Color bubbleColor = bubbleTypeColors[bubble->bubbleType];
        int isOutsideZone = bubble->isOutsideZone;
        
        if (isOutsideZone)
        {
            bubbleColor = ColorLerp(bubbleColor, COLOR_FROM_HEX(0xff4433), blink);
        }
        // lazy: encode id into rgb values
        bubbleColor.r = bubbleColor.r - i * 8 % 32;
        bubbleColor.g = bubbleColor.g - (i * 8 / 32) % 32;
        bubbleColor.b = bubbleColor.b - (i * 8 / 512) % 32;
        bubbleColor.a = 255;
        DrawModel(bubbleModel, position, bubble->radius + bubbleExtraRadius, bubbleColor);
    }
    rlPopMatrix();
}

void PlayfieldDrawRange(Playfield *playfield, Camera3D camera)
{
    Color rangeLimitColor = COLOR_FROM_HEX(0xff4400);
    int divides = 10;
    float divWidth = playfield->fieldSize.x / divides;
    for (int i = 0; i < divides; i+=2)
    {
        float x = i * divWidth - playfield->fieldSize.x * 0.5f + divWidth * 1.0f;
        DrawCube((Vector3){x, playfield->minHeight - playfield->fieldSize.y * 0.5f, 0}, divWidth, 5, 5, rangeLimitColor);
        DrawCube((Vector3){x, playfield->maxHeight - playfield->fieldSize.y * 0.5f, 0}, divWidth, 5, 5, rangeLimitColor);
    }
}

void PlayfieldFullDraw(Playfield *playfield, Camera3D camera)
{
    BeginDrawing();

    ClearBackground(COLOR_FROM_HEX(0x4488cc));

    BeginTextureMode(bubbleFieldTexture);
    rlSetClipPlanes(-128.0f, 128.0f);
    BeginMode3D(camera);

    ClearBackground(BLANK);
    PlayfieldDrawBubbles(playfield, camera);
    EndMode3D();
    EndTextureMode();

    float outlineSize = 1.0f;
    float outlineColor[4] = { 1.0f, 1.0f, 1.0f, 1.0f };
    float textureSize[2] = { (float)bubbleFieldTexture.texture.width, (float)bubbleFieldTexture.texture.height };

    SetShaderValue(bubbleOutlineShader, outlineSizeLoc, &outlineSize, SHADER_UNIFORM_FLOAT);
    SetShaderValue(bubbleOutlineShader, outlineColorLoc, outlineColor, SHADER_UNIFORM_VEC4);
    SetShaderValue(bubbleOutlineShader, textureSizeLoc, textureSize, SHADER_UNIFORM_VEC2);

    rlDisableDepthMask();
    BeginShaderMode(bubbleOutlineShader);
    DrawTexturePro(bubbleFieldTexture.texture, (Rectangle){0, 0, (float)bubbleFieldTexture.texture.width, 
        -(float)bubbleFieldTexture.texture.height}, (Rectangle){0, 0, (float)GetScreenWidth(), (float)GetScreenHeight()}, 
        (Vector2){0, 0}, 0.0f, WHITE);
    EndShaderMode();
    rlEnableDepthMask();

    BeginMode3D(camera);
    PlayfieldDrawRange(playfield, camera);
    EndMode3D();
}

void UpdateSceneGameOver()
{
    if (IsKeyPressed(KEY_ENTER))
    {
        nextScene = GAME_SCENE_MENU;
    }
    // Draw
    //----------------------------------------------------------------------------------

    BeginDrawing();

    ClearBackground(COLOR_FROM_HEX(0x4488cc));

    PlayfieldFullDraw(&playfield, camera);

    DrawText("Game Over", 30, 10, 40, WHITE);
    DrawText("Press [ENTER] to restart", 30, 50, 20, WHITE);

    PlayfieldScores scores = CalculatePlayfieldScores(&playfield);
    DrawText(TextFormat("Final Score: %d", scores.score), 30, 90, 20, WHITE);

    if (Button("Restart", (Vector2){30, 130}, (Vector2){100, 50}))
    {
        nextScene = GAME_SCENE_MENU;
    }
    EndDrawing();
}

void UpdateScenePlay()
{
    if (IsKeyPressed(KEY_ESCAPE))
    {
        nextScene = GAME_SCENE_MENU;
    }

    if (IsMouseButtonReleased(MOUSE_LEFT_BUTTON) && playfield.strikes < MAX_STRIKES)
    {
        Vector2 pos = PlayfieldScreenToSpawnPosition(&playfield, camera, 
            GetMousePosition());
        if (pos.y >= 0.0f && pos.y <= playfield.fieldSize.y
            && pos.x >= 0.0f && pos.x <= playfield.fieldSize.x)
        {
            PlayfieldTryAddBubble(&playfield, pos);
        }
    }
    
    while (gameTime.fixedTime < gameTime.time)
    {
        gameTime.fixedTime += gameTime.fixedDeltaTime;
        PlayfieldFixedUpdate(&playfield);
    }

    // Draw
    //----------------------------------------------------------------------------------

    PlayfieldFullDraw(&playfield, camera);

    PlayfieldScores scores = CalculatePlayfieldScores(&playfield);
    DrawText(TextFormat("Score: %d", scores.score), 10, 10, 20, WHITE);
    DrawText(TextFormat("Bubbles: %d", scores.bubbleCount), 10, 40, 20, WHITE);
    DrawText(TextFormat("Spawned: %d", playfield.spawnedBubbleCount), 10, 70, 20, WHITE);
    DrawText(TextFormat("Outside: %d", scores.outsideBubbleCount), 10, 100, 20, WHITE);
    DrawText(TextFormat("Strikes: %d", playfield.strikes), 10, 130, 20, WHITE);
    DrawText(TextFormat("Gameover in: %.1f", GAMEOVER_COOLDOWN - playfield.gameoverCooldown), 10, 160, 20, WHITE);

    EndDrawing();
}

void UpdateSceneMenu()
{
    // Draw
    //----------------------------------------------------------------------------------

    BeginDrawing();

    ClearBackground(COLOR_FROM_HEX(0x4488cc));

    DrawText("Bubble Pop", GetScreenWidth() / 2 - MeasureText("Bubble Pop", 40) / 2, 10, 40, WHITE);
    
    if (Button("Play easy", (Vector2){GetScreenWidth() / 2 - 100, 100}, (Vector2){200, 50}))
    {
        gameDifficulty = GAME_DIFFICULTY_EASY;
        nextScene = GAME_SCENE_PLAY;
    }

    if (Button("Play normal", (Vector2){GetScreenWidth() / 2 - 100, 160}, (Vector2){200, 50}))
    {
        gameDifficulty = GAME_DIFFICULTY_NORMAL;
        nextScene = GAME_SCENE_PLAY;
    }

    if (Button("Play hard", (Vector2){GetScreenWidth() / 2 - 100, 220}, (Vector2){200, 50}))
    {
        gameDifficulty = GAME_DIFFICULTY_HARD;
        nextScene = GAME_SCENE_PLAY;
    }

    EndDrawing();
}

int main(void)
{
    // Initialization
    //--------------------------------------------------------------------------------------
    const int screenWidth = 800;
    const int screenHeight = 450;

    InitWindow(screenWidth, screenHeight, "GGJ25 - Bubble Pop");

    SetTargetFPS(60);               // Set our game to run at 60 frames-per-second
    //--------------------------------------------------------------------------------------

    TraceLog(LOG_INFO, "loading shaders");
    bubbleOutlineShader = LoadShader(0, "data/bubble_outline.fs");
    // Get shader locations
    outlineSizeLoc = GetShaderLocation(bubbleOutlineShader, "outlineSize");
    outlineColorLoc = GetShaderLocation(bubbleOutlineShader, "outlineColor");
    textureSizeLoc = GetShaderLocation(bubbleOutlineShader, "textureSize");



    bubbleModel = LoadModelFromMesh(GenMeshSphere(1.0f, 4, 24));
    // Main game loop
    while (!WindowShouldClose())    // Detect window close button or ESC key
    {
        if (bubbleFieldTexture.texture.width != GetScreenWidth() || bubbleFieldTexture.texture.height != GetScreenHeight())
        {
            UnloadRenderTexture(bubbleFieldTexture);
            bubbleFieldTexture = LoadRenderTexture(GetScreenWidth(), GetScreenHeight());
        }

        float dt = GetFrameTime();
        // clamp dt to prevent large time steps, e.g. when browser tab is inactive
        if (dt > 0.2f) dt = 0.2f;
        gameTime.time += dt;
        gameTime.deltaTime = dt;

        switch (currentScene)
        {
            default: UpdateSceneMenu(); break;
            case GAME_SCENE_PLAY: UpdateScenePlay(); break;
            case GAME_SCENE_GAMEOVER: UpdateSceneGameOver(); break;
        }

        switch (nextScene)
        {
            case GAME_SCENE_NONE: break;
            default: currentScene = nextScene; break;
            case GAME_SCENE_PLAY: 
                playfield = (Playfield){
                    .fieldSize = {250, 300},
                    .waterLevel = 200.0f,
                    .minHeight = 90.0f,
                    .maxHeight = 280.0f,
                };
                currentScene = GAME_SCENE_PLAY; break;
        }

        nextScene = GAME_SCENE_NONE;
    }

    // De-Initialization
    //--------------------------------------------------------------------------------------
    CloseWindow();        // Close window and OpenGL context
    //--------------------------------------------------------------------------------------

    return 0;
}